@@ -1,8 +1,8 @@
 import os
 from glob import glob
 import os.path as op
+from pathlib import Path
 import ruamel.yaml
-import yaml
 
 from ..consts import file_operation_modes
 
@@ -13,15 +13,13 @@ from ..organize import (
     create_unique_filenames_from_metadata,
     get_obj_id,
     populate_dataset_yml,
+    detect_link_type,
 )
-from ..utils import find_files, on_windows
+from ..utils import find_files, on_windows, yaml_load
 from ..pynwb_utils import copy_nwb_file, get_object_id
 import pytest
 
 
-yaml_ruamel = ruamel.yaml.YAML()  # defaults to round-trip if no parameters given
-
-
 def test_sanitize_value():
     # . is not sanitized in extension but elsewhere
     assert _sanitize_value("_.ext", "extension") == "-.ext"
@@ -34,7 +32,7 @@ def test_populate_dataset_yml(tmpdir):
 
     def c():  # shortcut
         with open(path) as f:
-            return yaml.safe_load(f)
+            return yaml_load(f, typ="safe")
 
     path.write("")
     populate_dataset_yml(str(path), [])  # doesn't crash
@@ -51,7 +49,7 @@ def test_populate_dataset_yml(tmpdir):
         {"age": 2, "cell_id": "2", "tissue_sample_id": 1, "sex": "F"},
     ]
 
-    # even though we use ruyaml for manipulation, we should assure it is readable
+    # even though we use ruamel for manipulation, we should assure it is readable
     # by regular yaml
     populate_dataset_yml(str(path), metadata)
     assert c() == {
@@ -63,9 +61,9 @@ def test_populate_dataset_yml(tmpdir):
     }
 
     # and if we set units and redo -- years should stay unchanged, while other fields change
-    m = yaml_ruamel.load(path.read())
+    m = yaml_load(path.read())
     m["age"]["units"] = "years"
-    yaml_ruamel.dump(m, open(path, "w"))
+    ruamel.yaml.YAML().dump(m, open(path, "w"))
 
     populate_dataset_yml(str(path), metadata[:1])
     assert c() == {
@@ -118,6 +116,31 @@ def test_organize_nwb_test_data(nwb_test_data, tmpdir, clirunner, mode):
         nwb_test_data = op.relpath(nwb_test_data, cwd)
         outdir = op.relpath(outdir, cwd)
 
+    src = Path(tmpdir, "src")
+    src.touch()
+    dest = Path(tmpdir, "dest")
+    try:
+        dest.symlink_to(src)
+    except OSError:
+        symlinks_work = False
+    else:
+        symlinks_work = True
+    try:
+        dest.unlink()
+    except FileNotFoundError:
+        pass
+    try:
+        os.link(src, dest)
+    except OSError:
+        hardlinks_work = False
+    else:
+        hardlinks_work = True
+
+    if mode in ("simulate", "symlink") and not symlinks_work:
+        pytest.skip("Symlinks not supported")
+    elif mode == "hardlink" and not hardlinks_work:
+        pytest.skip("Hard links not supported")
+
     input_files = op.join(nwb_test_data, "v2.0.1")
 
     cmd = ["-d", outdir, "--files-mode", mode, input_files]
@@ -126,7 +149,7 @@ def test_organize_nwb_test_data(nwb_test_data, tmpdir, clirunner, mode):
     # with @map_to_click_exceptions we loose original str of message somehow
     # although it is shown to the user - checked. TODO - figure it out
     # assert "not containing all" in str(r.exc_info[1])
-    assert r.exit_code != 0, f"Must have aborted since many files lack subject_id"
+    assert r.exit_code != 0, "Must have aborted since many files lack subject_id"
     assert not glob(op.join(outdir, "*")), "no files should have been populated"
 
     r = clirunner.invoke(organize, cmd + ["--invalid", "warn"])
@@ -134,7 +157,7 @@ def test_organize_nwb_test_data(nwb_test_data, tmpdir, clirunner, mode):
     # this beast doesn't capture our logs ATM so cannot check anything there.
     # At the end we endup only with dandiset.yaml and a single file
     produced_paths = sorted(find_files(".*", paths=outdir))
-    produced_nwb_paths = sorted(find_files(".nwb$", paths=outdir))
+    produced_nwb_paths = sorted(find_files(r"\.nwb\Z", paths=outdir))
     produced_relpaths = [op.relpath(p, outdir) for p in produced_paths]
     if mode == "dry":
         assert produced_relpaths == []
@@ -149,7 +172,7 @@ def test_organize_nwb_test_data(nwb_test_data, tmpdir, clirunner, mode):
 
     if mode == "simulate":
         assert all((op.isabs(p) != relative) for p in produced_paths)
-    elif mode == "symlink":
+    elif mode == "symlink" or (mode == "auto" and symlinks_work):
         assert all(op.islink(p) for p in produced_nwb_paths)
     else:
         assert not any(op.islink(p) for p in produced_paths)
@@ -186,24 +209,45 @@ def test_ambiguos_probe1():
     metadata_ = create_unique_filenames_from_metadata(metadata)
     assert metadata_ != metadata
     assert [m["dandi_path"] for m in metadata_] == [
-        "sub-1/sub-1.nwb",
-        "sub-1/sub-1_probe-1_mod.nwb",
-        "sub-1/sub-1_probe-2_mod.nwb",
+        op.join("sub-1", "sub-1.nwb"),
+        op.join("sub-1", "sub-1_probe-1_mod.nwb"),
+        op.join("sub-1", "sub-1_probe-2_mod.nwb"),
     ]
     # if modalities is present but different -- no probe for _mod2
     metadata[0]["modalities"] = ["mod2"]
     metadata_ = create_unique_filenames_from_metadata(metadata)
     assert [m["dandi_path"] for m in metadata_] == [
-        "sub-1/sub-1_mod2.nwb",
-        "sub-1/sub-1_probe-1_mod.nwb",
-        "sub-1/sub-1_probe-2_mod.nwb",
+        op.join("sub-1", "sub-1_mod2.nwb"),
+        op.join("sub-1", "sub-1_probe-1_mod.nwb"),
+        op.join("sub-1", "sub-1_probe-2_mod.nwb"),
     ]
 
     # but if modalities is same -- we would get probes listed
     metadata[0]["modalities"] = ["mod"]
     metadata_ = create_unique_filenames_from_metadata(metadata)
     assert [m["dandi_path"] for m in metadata_] == [
-        "sub-1/sub-1_probe-1+2_mod.nwb",
-        "sub-1/sub-1_probe-1_mod.nwb",
-        "sub-1/sub-1_probe-2_mod.nwb",
+        op.join("sub-1", "sub-1_probe-1+2_mod.nwb"),
+        op.join("sub-1", "sub-1_probe-1_mod.nwb"),
+        op.join("sub-1", "sub-1_probe-2_mod.nwb"),
     ]
+
+
+@pytest.mark.parametrize(
+    "sym_success,hard_success,result",
+    [
+        (True, True, "symlink"),
+        (True, False, "symlink"),
+        (False, True, "hardlink"),
+        (False, False, "copy"),
+    ],
+)
+def test_detect_link_type(monkeypatch, tmp_path, sym_success, hard_success, result):
+    def succeed_link(src, dest):
+        pass
+
+    def error_link(src, dest):
+        raise OSError("Operation failed")
+
+    monkeypatch.setattr(os, "symlink", succeed_link if sym_success else error_link)
+    monkeypatch.setattr(os, "link", succeed_link if hard_success else error_link)
+    assert detect_link_type(tmp_path) == result
