@@ -1,5 +1,6 @@
 import datetime
 import inspect
+import io
 import itertools
 import os
 import os.path as op
@@ -10,8 +11,15 @@ import sys
 
 from pathlib import Path
 
-if sys.version_info[:2] < (3, 7):
-    import dateutil.parser
+import requests
+import ruamel.yaml
+from semantic_version import Version
+
+from . import __version__
+from .consts import dandi_instance, known_instances, known_instances_rev
+from .exceptions import BadCliVersionError, CliVersionTooOldError
+
+import dateutil.parser
 
 #
 # Additional handlers
@@ -141,7 +149,11 @@ def ensure_strtime(t, isoformat=True):
 
 def fromisoformat(t, impl=None):
     if impl is None:
-        impl = "datetime" if sys.version_info[:2] >= (3, 7) else "dateutil"
+        # datetime parser "does not support parsing arbitrary ISO 8601 strings"
+        # https://docs.python.org/3/library/datetime.html
+        # In particular it does not parse time zone suffix which was recently
+        # introduced into datetime's provided by API
+        impl = "dateutil"
     if impl == "datetime":
         return datetime.datetime.fromisoformat(t)
     elif impl == "dateutil":
@@ -157,7 +169,6 @@ def ensure_datetime(t, strip_tzinfo=False, tz=None):
 
     epoch time assumed to be local (not utc)
     """
-    t_orig = t
     if isinstance(t, datetime.datetime):
         pass
     elif isinstance(t, (int, float)):
@@ -203,6 +214,39 @@ def updated(d, update):
     return d
 
 
+def remap_dict(rec, revmapping):
+    """Remap nested dicts according to mapping
+
+    Parameters
+    ----------
+    revmapping: dict
+      (to, from)
+
+    TODO: document and test more
+    """
+    out = {}
+
+    def split(path):
+        # map path from key.subkey if given in a string (not tuple) form
+        return path.split(".") if isinstance(path, str) else path
+
+    for to, from_ in revmapping.items():
+        in_v = rec
+        for p in split(from_):
+            if p not in in_v:
+                continue  # it is not there -- cannot map
+            in_v = in_v[p]
+
+        # and now set
+        out_v = out
+        t_split = split(to)
+        for p in t_split[:-1]:
+            out_v[p] = out_v.get(p, {})  # container for the next field
+            out_v = out_v[p]
+        out_v[t_split[-1]] = in_v  # and the last one gets the in_v
+    return out
+
+
 #
 # Paths and files
 #
@@ -319,7 +363,8 @@ def move_file(src, dst):
 def find_dandi_files(paths):
     """Adapter to find_files to find files of interest to dandi project
     """
-    yield from find_files("(dandiset\.yaml|\.nwb)$", paths)
+    sep = re.escape(os.sep)
+    yield from find_files(rf"((^|{sep})dandiset\.yaml|\.nwb)\Z", paths)
 
 
 def find_parent_directory_containing(filename, path=None):
@@ -348,9 +393,31 @@ def yaml_dump(rec):
     to assure proper formatting on versions of pyyaml before
     5.1: https://github.com/yaml/pyyaml/pull/256
     """
-    import yaml
+    yaml = ruamel.yaml.YAML(typ="safe")
+    yaml.default_flow_style = False
+    out = io.StringIO()
+    yaml.dump(rec, out)
+    return out.getvalue()
 
-    return yaml.safe_dump(rec, default_flow_style=False)
+
+def yaml_load(f, typ=None):
+    """
+    Load YAML source from a file or string.
+
+    Parameters
+    ----------
+    f: str or IO[str]
+      The YAML source to load
+    typ: str, optional
+      The value of `typ` to pass to `ruamel.yaml.YAML`.  May be "rt" (default),
+      "safe", "unsafe", or "base"
+
+    Returns
+    -------
+    Any
+      The parsed YAML value
+    """
+    return ruamel.yaml.YAML(typ=typ).load(f)
 
 
 #
@@ -408,14 +475,14 @@ def safe_call(func, path, default=None):
         return default
 
 
-def shortened_repr(value, l=30):
+def shortened_repr(value, length=30):
     try:
         if hasattr(value, "__repr__") and (value.__repr__ is not object.__repr__):
             value_repr = repr(value)
-            if not value_repr.startswith("<") and len(value_repr) > l:
+            if not value_repr.startswith("<") and len(value_repr) > length:
                 value_repr = "<<%s++%d chars++%s>>" % (
-                    value_repr[: l - 16],
-                    len(value_repr) - (l - 16 + 4),
+                    value_repr[: length - 16],
+                    len(value_repr) - (length - 16 + 4),
                     value_repr[-4:],
                 )
             elif (
@@ -426,7 +493,7 @@ def shortened_repr(value, l=30):
                 raise ValueError("I hate those useless long reprs")
         else:
             raise ValueError("gimme class")
-    except Exception as e:
+    except Exception:
         value_repr = "<%s>" % value.__class__.__name__.split(".")[-1]
     return value_repr
 
@@ -480,3 +547,51 @@ def delayed(*args, **kwargs):
     import joblib
 
     return joblib.delayed(*args, **kwargs)
+
+
+def get_instance(dandi_instance_id):
+    if dandi_instance_id.lower().startswith(("http://", "https://")):
+        redirector_url = dandi_instance_id
+        dandi_id = known_instances_rev.get(redirector_url)
+        if dandi_id is not None:
+            instance = known_instances[dandi_id]
+        else:
+            instance = None
+    else:
+        instance = known_instances[dandi_instance_id]
+        redirector_url = instance.redirector
+        if redirector_url is None:
+            return instance
+    try:
+        r = requests.get(redirector_url.rstrip("/") + "/server-info")
+        r.raise_for_status()
+    except Exception as e:
+        lgr.warning("Request to %s failed (%s)", redirector_url, str(e))
+        if instance is not None:
+            lgr.warning("Using hard-coded URLs")
+            return instance
+        else:
+            raise RuntimeError(
+                f"Could not retrieve server info from {redirector_url},"
+                " and client does not recognize URL"
+            )
+    server_info = r.json()
+    try:
+        minversion = Version(server_info["cli-minimal-version"])
+        bad_versions = [Version(v) for v in server_info["cli-bad-versions"]]
+    except ValueError as e:
+        raise ValueError(
+            f"{redirector_url} returned an incorrectly formatted version;"
+            f" please contact that server's administrators: {e}"
+        )
+    our_version = Version(__version__)
+    if our_version < minversion:
+        raise CliVersionTooOldError(our_version, minversion, bad_versions)
+    if our_version in bad_versions:
+        raise BadCliVersionError(our_version, minversion, bad_versions)
+    return dandi_instance(
+        girder=server_info["services"]["girder"]["url"],
+        gui=server_info["services"]["webui"]["url"],
+        redirector=redirector_url,
+        api=server_info["services"]["api"]["url"],
+    )
